from uptrain.core.lib.helper_funcs import extract_data_points_from_batch
from uptrain.core.classes.statistics import AbstractStatistic
from uptrain.core.classes.distances import DistanceResolver
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Statistic


class Distance(AbstractStatistic):
    dashboard_name = "distance"
    anomaly_type = Statistic.DISTANCE

    def __init__(self, fw, check):
        self.log_handler = fw.log_handler
        self.log_handler.add_writer(self.dashboard_name)
        self.measurable = MeasurableResolver(check["measurable_args"]).resolve(fw)
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.item_counts = {}
        self.feats_dictn = {}
        self.reference = check["reference"]
        self.distance_types = check["distance_types"]
        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]

    def check(self, inputs, outputs=None, gts=None, extra_args={}):
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )

        for idx in range(len(aggregate_ids)):
            if aggregate_ids[idx] not in self.item_counts:
                self.item_counts.update({aggregate_ids[idx]: 0})
            this_item_count = self.item_counts[aggregate_ids[idx]]
            this_val = extract_data_points_from_batch(vals, [idx])

            if this_item_count > 0:
                this_distances = dict(
                    zip(
                        self.distance_types,
                        [
                            x.compute_distance(
                                this_val, self.feats_dictn[aggregate_ids[idx]]
                            )
                            for x in self.dist_classes
                        ],
                    )
                )
                if self.reference == "running_diff":
                    self.feats_dictn[aggregate_ids[idx]] = this_val
                for distance_type in self.distance_types:
                    plot_name = (
                        distance_type
                        + "_"
                        + str(self.reference)
                        # + self.measurable.col_name()
                        # + " "
                        # + self.aggregate_measurable.col_name()
                    )
                    self.log_handler.add_scalars(
                        self.dashboard_name + "_" + plot_name,
                        {str(aggregate_ids[idx]): this_distances[distance_type][0]},
                        this_item_count,
                        self.dashboard_name,
                    )
            else:
                self.feats_dictn.update({aggregate_ids[idx]: this_val})
            self.item_counts[aggregate_ids[idx]] += 1



import numpy as np


class NormRatio:
    def compute_distance(self, base, ref) -> float:
        base = np.array(base)
        ref = np.array(ref)
        base_norm = np.linalg.norm(base, axis=1)
        ref_norm = np.linalg.norm(ref, axis=1)
        return base_norm / np.maximum(ref_norm, 1e-6)



import numpy as np

from uptrain.core.classes.anomalies import AbstractAnomaly
from uptrain.core.classes.signals import SignalManager
from uptrain.constants import Anomaly


class EdgeCase(AbstractAnomaly):
    dashboard_name = "edge_cases"
    anomaly_type = Anomaly.EDGE_CASE

    def __init__(self, fw, signal_formulae):
        self.log_handler = fw.log_handler
        self.log_handler.add_writer(self.dashboard_name)
        self.signal_manager = SignalManager()
        self.signal_manager.add_signal_formulae(signal_formulae)
        self.num_preds = 0
        self.num_selected = 0

    def check(self, inputs, outputs, gts=None, extra_args={}):
        return

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        is_interesting = self.signal_manager.evaluate_signal(
            inputs, outputs, gts=gts, extra_args=extra_args
        )
        self.num_preds += len(is_interesting)
        self.num_selected += sum(is_interesting)
        self.log_handler.add_scalars(
            "num_edge_cases",
            {"num_edge_cases": self.num_selected},
            self.num_preds,
            self.dashboard_name,
        )
        return is_interesting

    def need_ground_truth(self):
        return False



import streamlit as st
import pandas as pd
from glob import glob
import os
import sys
import plotly.graph_objects as go
import numpy as np
import json
import plotly.express as px


def return_plotly_fig(y_axis, x_axis="Num predictions", x_log=False, y_log=False):
    fig = go.Figure()
    fig.update_xaxes(title_text=x_axis)
    fig.update_yaxes(title_text=y_axis)
    if x_log:
        fig.update_xaxes(type="log")
    if y_log:
        fig.update_yaxes(type="log")
    return fig


st.set_page_config(
    page_title="UpTrain AI Dashboard",
    layout="wide",
    page_icon='https://github.com/uptrain-ai/uptrain/raw/dashboard/uptrain/core/classes/logging/uptrain_logo_icon.png'
)
st.title("UpTrain AI Live Dashboard")
st_style = """
            <style>
            footer {visibility: hidden;}
            </style>
           """
st.markdown(st_style, unsafe_allow_html=True)


def plot_dashboard(dashboard_name):
    if st.sidebar.checkbox(f"Show dashboard: {dashboard_name}", value=True):
        sub_dirs = [path[0] for path in os.walk(os.path.join(log_folder, dashboard_name))]
        for sub_dir in sub_dirs:
            sub_dir_split = sub_dir.split("/")

            ######### Showing Alerts ###########

            if sub_dir_split[-1] == "alerts":
                files = [
                            file
                            for path, _, _ in os.walk(sub_dir)
                            for file in glob(os.path.join(path, "*.json"))
                        ]
                for i, file in enumerate(files):
                    alert_name = file.split("/")[-1].split(".")[0]
                    f = open(file)
                    alert = json.load(f)
                    st.subheader(alert_name)
                    st.markdown("##### " + alert)
                    st.markdown("""---""")

                st.sidebar.markdown("""---""")

            ######### Line Plots ###########

            elif sub_dir_split[-2] == "line_plots":
                plot_name = sub_dir_split[-1]

                if st.sidebar.checkbox(f"Line-plot for {plot_name}", value=True):
                    st.markdown(f"### Line chart for {plot_name}")

                    # Getting the list of all csv files in streamlit logs
                    csv_files = [
                        file
                        for path, _, _ in os.walk(sub_dir)
                        for file in glob(os.path.join(path, "*.csv"))
                    ]

                    scol1, scol2 = st.columns(2)
                    with scol1:
                        x_log = st.checkbox(
                            "log x", help="Plot x-axis in log-scale", key=plot_name + "x"
                        )
                    with scol2:
                        y_log = st.checkbox(
                            "log y", help="Plot y-axis in log-scale", key=plot_name + "y"
                        )
                    fig = return_plotly_fig(plot_name, x_log=x_log, y_log=y_log)
                    for i, csv_file in enumerate(csv_files):
                        # Reading the csv file
                        df = pd.read_csv(csv_file)

                        # Getting plot_id
                        plot_id = csv_file.split("/")[-1].split(".")[0]
                        fig = fig.add_trace(
                            go.Scatter(
                                x=df["count"],
                                y=df[plot_id],
                                name=str(i) + ", " + plot_id,
                            )
                        )
                    st.plotly_chart(fig)
                    st.markdown("""---""")    

                st.sidebar.markdown("""---""")

            ######### Plotting histograms ###########

            elif sub_dir_split[-2] == "histograms":
                plot_name = sub_dir_split[-1]

                if plot_name != "umap_and_clusters":
                    if st.sidebar.checkbox(f"Histogram for {plot_name}", value=True):
                        st.markdown(f"### Histogram for {plot_name}")

                        # Getting the list of all files in streamlit logs
                        files = [
                            file
                            for path, _, _ in os.walk(sub_dir)
                            for file in glob(os.path.join(path, "*.json"))
                        ]

                        for i, file in enumerate(files):
                            count = file.split("/")[-1].split(".")[0]
                            if int(count) < 0:
                                with open(file, encoding='utf-8') as f:
                                    data = json.loads(f.read())
                                if isinstance(data, dict):
                                    fig = go.Figure()
                                    for key in data.keys():
                                        fig = fig.add_trace(go.Histogram(x=data[key], name=key))
                                else:
                                    fig = go.Figure(data=[go.Histogram(x=data)])
                                st.plotly_chart(fig)
                                st.markdown("""---""")   
                            else:
                                if st.checkbox(f"{plot_name} histogram for count {count}"):
                                    f = open(file)
                                    data = json.load(f)
                                    fig = go.Figure(data=[go.Histogram(x=data, name=count)])
                                    st.plotly_chart(fig)
                                    st.markdown("""---""")   
                                
                else:
                    if st.sidebar.checkbox(f"{plot_name}"):
                        st.markdown(f"### {plot_name}")

                        # Getting the list of all files in streamlit logs
                        files = [
                            file
                            for path, _, _ in os.walk(sub_dir)
                            for file in glob(os.path.join(path, "*.json"))
                        ]

                        for i, file in enumerate(files):
                            count = file.split("/")[-1].split(".")[0]
                            if st.checkbox(f"{plot_name} UMAP and Clusters for count {count}"):
                                with open(file, encoding='utf-8') as f:
                                    data = json.loads(f.read())
                                arr = np.array(data["umap"])
                                clusters = data["clusters"]
                                x = arr[:, 0]
                                y = arr[:, 1]
                                if arr.shape[1] == 2:
                                    fig = px.scatter(x=x, y=y, color=clusters)
                                elif arr.shape[1] == 3:
                                    z = arr[:, 2]
                                    fig = px.scatter_3d(x=x, y=y, z=z, color=clusters)
                                else:
                                    raise ("Umap dimension not 2D or 3D.")
                                st.plotly_chart(fig, use_container_width=True)
                                # st.write(
                                #     f"Number of clusters for count {count}: {len(set(clusters))-1}"
                                # )
                                st.markdown("""---""")   
                st.sidebar.markdown("""---""")

            ######### Plotting Bar Graphs ###########

            elif sub_dir_split[-2] == "bar_graphs":
                plot_name = sub_dir_split[-1]

                if st.sidebar.checkbox(f"Bar graph for {plot_name}", value=True):
                    st.markdown(f"### Bar graph for {plot_name}")

                    # Getting the list of all files in streamlit logs
                    files = [
                        file
                        for path, _, _ in os.walk(sub_dir)
                        for file in glob(os.path.join(path, "*.json"))
                    ]

                    for i, file in enumerate(files):
                        count = file.split("/")[-1].split(".")[0]
                        if int(count) < 0:
                            with open(file, encoding='utf-8') as f:
                                data = json.loads(f.read())
                            fig = go.Figure()
                            for bar_name in data:
                                bar_dict = data[bar_name]
                                keys, values = zip(*bar_dict.items())
                                fig = fig.add_trace(
                                    go.Bar(
                                        x=list(keys),
                                        y=list(values),
                                        name=bar_name,
                                    )
                                )
                            st.plotly_chart(fig)
                            st.markdown("""---""")   
                        else:
                            if st.checkbox(f"{plot_name} Bar graph for count {count}"):
                                with open(file, encoding='utf-8') as f:
                                    data = json.loads(f.read())
                                fig = go.Figure()
                                for bar_name in data:
                                    bar_dict = data[bar_name]
                                    keys, values = zip(*bar_dict.items())
                                    fig = fig.add_trace(
                                        go.Bar(
                                            x=list(keys),
                                            y=list(values),
                                            name=bar_name,
                                        )
                                    )
                                st.plotly_chart(fig)
                                st.markdown("""---""")   
                st.sidebar.markdown("""---""")



# Getting the streamlit log folder
log_folder = sys.argv[1]
st.sidebar.title("Select dashboards to view")
dashboard_names = next(os.walk(log_folder))[1]

for dashboard_name in dashboard_names:
    plot_dashboard(dashboard_name)


class AbstractAnomaly:
    anomaly_type = None

    def check(self, inputs, outputs, gts=None, extra_args={}):
        raise Exception("Should be defined for each class")

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        raise Exception("Should be defined for each class")

    def need_ground_truth(self):
        return False



from .abstract_anomaly import AbstractAnomaly
from .concept_drift import ConceptDrift
from .custom_anomaly import CustomAnomaly
from .data_drift import DataDrift
from .edge_case import EdgeCase
from .recommendation_bias import RecommendationBias
from .data_integrity import DataIntegrity



import numpy as np
import copy

from uptrain.core.classes.anomalies import AbstractAnomaly
from uptrain.constants import Anomaly
from uptrain.core.lib.helper_funcs import read_json, cluster_and_plot_data
from uptrain.core.classes.measurables import MeasurableResolver


class DataDrift(AbstractAnomaly):
    dashboard_name = "data_drift"
    anomaly_type = Anomaly.DATA_DRIFT
    is_embedding = None
    mode = None
    NUM_BUCKETS = 20
    INITIAL_SKIP = 2000

    def __init__(self, fw, check, is_embedding=None):
        self.measurable = None
        self.measurable = MeasurableResolver(check["measurable_args"]).resolve(fw)
        self.is_embedding = is_embedding
        if self.is_embedding == None:
            self.embedding_child_class = DataDrift(fw, check, is_embedding=True)
            self.scalar_child_class = DataDrift(fw, check, is_embedding=False)
        else:
            self.log_handler = fw.log_handler
            self.reference_dataset = check["reference_dataset"]
            self.cluster_plot_func = fw.dataset_handler.cluster_plot_func
            self.save_edge_cases = check.get("save_edge_cases", True)
            self.NUM_BUCKETS = check.get("num_buckets", self.NUM_BUCKETS)
            self.INITIAL_SKIP = check.get("initial_skip", self.INITIAL_SKIP)
            self.count = 0
            self.prod_dist_counts_arr = []
            if self.is_embedding:
                self.plot_name = "Embeddings_" + self.measurable.col_name()
                self.emd_threshold = check.get("emd_threshold", 1)
            else:
                self.plot_name = "Scalar_" + self.measurable.col_name()
            self.bucket_reference_dataset()

    def need_ground_truth(self):
        return False

    def check(self, inputs, outputs, gts=None, extra_args={}):
        if self.is_embedding == None:
            if self.mode == "check_both":
                self.embedding_child_class.check(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
                self.scalar_child_class.check(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
            elif self.mode == "check_scalar_only":
                self.scalar_child_class.check(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
            elif self.mode == "check_embeddings_only":
                self.scalar_child_class.check(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
            else:
                feats = self.measurable.compute_and_log(
                    inputs=inputs, outputs=outputs, gts=gts, extra=extra_args
                )
                if len(list(feats.shape)) > 2:
                    self.mode = "check_both"
                else:
                    self.mode = "check_scalar_only"
                self.check(inputs, outputs, gts=gts, extra_args=extra_args)
        else:
            # if (self.count == 0) and not(self.is_embedding and (self.cluster_plot_func is not None)):
            if self.count == 0:
                self.log_handler.add_writer(self.dashboard_name)

            self.count += len(extra_args["id"])

            self.feats = self.measurable.compute_and_log(
                inputs=inputs, outputs=outputs, gts=gts, extra=extra_args
            )
            feats_shape = list(self.feats.shape)
            feats_shape.insert(1, 1)
            if len(feats_shape) == 2:
                feats_shape.insert(2, 1)
            self.feats = np.reshape(self.feats, tuple(feats_shape))

            if self.is_embedding:
                self.feats = self.feats / np.expand_dims(self.max_along_axis, 0)
                selected_cluster = np.argmin(
                    np.sum(
                        np.abs(self.clusters[0] - self.feats),
                        axis=tuple(range(2, len(feats_shape))),
                    ),
                    axis=1,
                )
                for clus in selected_cluster:
                    self.prod_dist_counts[0][clus] += 1
                self.this_datapoint_cluster = selected_cluster
                self.prod_dist_counts_arr.append(self.prod_dist_counts.copy())
            else:
                this_datapoint_cluster = []
                for idx in range(self.feats.shape[2]):
                    if isinstance(self.feats[0,0,idx], str):
                        try:
                            bucket_idx = np.array([
                                list(self.buckets[idx]).index(self.feats[x,0,idx]) for x in range(self.feats.shape[0])
                            ])
                        except:
                            # TODO: This logic is not completely tested yet. Contact us if you are facing issues
                            # If given data-point is not present -> add a new bucket
                            temp_buckets = list(self.buckets[idx])
                            num_added = 0

                            for x in range(self.feats.shape[0]):
                                if self.feats[x,0,idx] not in temp_buckets:
                                    temp_buckets.append(self.feats[x,0,idx])
                                    num_added += 1
                            
                            self.buckets[idx] = np.array(temp_buckets)
                            bucket_idx = np.array([
                                list(self.buckets[idx]).index(self.feats[x,0,idx]) for x in range(self.feats.shape[0])
                            ])
                    else:
                        bucket_idx = np.searchsorted(
                            self.buckets[idx], self.feats[:, :, idx]
                        )[:, 0]
                    for clus in bucket_idx:
                        self.prod_dist_counts[idx][clus] += 1
                    this_datapoint_cluster.append(bucket_idx)
                self.prod_dist_counts_arr.append(self.prod_dist_counts.copy())
                self.this_datapoint_cluster = np.array(this_datapoint_cluster)

            self.prod_dist = (
                self.prod_dist_counts_arr[-1]
                - self.prod_dist_counts_arr[
                    int(max(self.count - self.INITIAL_SKIP, 0) / len(extra_args["id"]))
                ]
            ) / min(self.count, self.INITIAL_SKIP)

            if self.ref_dist.shape[0] == 1:
                if self.is_embedding:
                    bar_graph_buckets = ["cluster_" + str(idx) for idx in range(len(list(np.reshape(self.clusters,-1))))]
                else:
                    bar_graph_buckets = list(np.reshape(self.clusters,-1))

                self.log_handler.add_bar_graphs(
                    self.plot_name,
                    {
                        "reference": dict(zip(bar_graph_buckets, list(np.reshape(np.array(self.ref_dist[0]), -1)))),
                        "production": dict(zip(bar_graph_buckets, list(np.reshape(np.array(self.prod_dist[0]), -1)))),
                    },
                    self.dashboard_name,
                )

            if self.count < self.INITIAL_SKIP:
                self.drift_detected = False
                return
            else:
                self.psis = np.zeros(self.ref_dist.shape[0])
                self.costs = np.zeros(self.ref_dist.shape[0])
                drift_detected = False
                for idx in range(self.ref_dist.shape[0]):
                    if self.is_embedding:
                        self.costs[idx] = self.estimate_earth_moving_cost(
                            self.prod_dist[idx], self.ref_dist[idx], self.clusters[idx]
                        )
                        drift_detected = drift_detected or (self.costs[idx] > self.emd_threshold)
                    else:
                        this_psi = sum(
                            [
                                (self.prod_dist[idx][jdx] - self.ref_dist[idx][jdx])
                                * np.log(
                                    max(self.prod_dist[idx][jdx], 0.0001)
                                    / self.ref_dist[idx][jdx]
                                )
                                for jdx in range(self.ref_dist.shape[1])
                            ]
                        )
                        self.psis[idx] = this_psi
                        drift_detected = drift_detected or (this_psi > 0.3)
                self.drift_detected = drift_detected
                if self.drift_detected:
                    alert = "Data Drift last detected at " + str(self.count) + " for Embeddings with Earth moving distance = " + str(float(self.costs[0]))
                    self.log_handler.add_alert(
                        "Data Drift Alert 🚨",
                        alert,
                        self.dashboard_name
                    )

            if self.is_embedding:
                dict_emc = dict(
                    zip(
                        [self.measurable.col_name() for x in range(self.costs.shape[0])],
                        [float(x) for x in list(self.costs)],
                    )
                )
                dict_emc.update({"threshold": self.emd_threshold})
                self.log_handler.add_scalars(
                    self.measurable.col_name() + " - earth_moving_costs_embedding",
                    dict_emc,
                    self.count,
                    self.dashboard_name,
                )
            else:
                dict_psi = dict(
                    zip(
                        [self.measurable.col_name() + "_" + str(x) for x in range(self.psis.shape[0])],
                        [float(x) for x in list(self.psis)],
                    )
                )
                dict_psi.update({"threshold": 0.3})
                self.log_handler.add_scalars(
                    self.measurable.col_name() + " - population_stability_index_scalar",
                    dict_psi,
                    self.count,
                    self.dashboard_name,
                )

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        if self.is_embedding == None:
            if self.mode == "check_both":
                arr1 = self.embedding_child_class.is_data_interesting(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
                arr2 = self.scalar_child_class.is_data_interesting(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
                return np.logical_or(np.array(arr1), np.array(arr2))
            elif self.mode == "check_scalar_only":
                return self.scalar_child_class.is_data_interesting(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
            elif self.mode == "check_embeddings_only":
                return self.scalar_child_class.is_data_interesting(
                    inputs, outputs, gts=gts, extra_args=extra_args
                )
            else:
                raise Exception("Mode for data drift should be set in check func")
        else:
            is_interesting = np.array([False] * len(extra_args["id"]))
            if self.save_edge_cases and self.drift_detected and not isinstance(self.feats[0,0,0], str):
                feats_shape = list(self.feats.shape)
                del feats_shape[1]
                self.feats = np.reshape(self.feats, tuple(feats_shape))
                for idx in range(self.ref_dist.shape[0]):
                    if self.is_embedding:
                        dist_from_cluster = (
                            np.sum(
                                np.abs(
                                    self.clusters[idx][self.this_datapoint_cluster]
                                    - self.feats
                                ),
                                axis=tuple(range(1, len(feats_shape))),
                            )
                            / self.cluster_vars[idx][self.this_datapoint_cluster]
                        )
                    else:
                        dist_from_cluster = np.abs(
                            np.reshape(
                                self.clusters[idx][self.this_datapoint_cluster[idx]], -1
                            )
                            - self.feats[:, idx]
                        ) / np.reshape(
                            self.cluster_vars[idx][self.this_datapoint_cluster[idx]], -1
                        )
                    for jdx in range(dist_from_cluster.shape[0]):
                        if dist_from_cluster[jdx] > 2:
                            if self.is_embedding:
                                min_dist_from_clusters = np.min(
                                    np.sum(
                                        np.abs(self.clusters[idx] - self.feats[jdx]),
                                        axis=tuple(range(1, len(feats_shape))),
                                    )
                                    / self.cluster_vars[idx]
                                )
                            else:
                                min_dist_from_clusters = np.min(
                                    np.abs(
                                        np.reshape(self.clusters[idx], -1)
                                        - self.feats[jdx, idx]
                                    )
                                    / np.reshape(self.cluster_vars[idx], -1)
                                )
                            is_interesting[jdx] = is_interesting[jdx] or (
                                min_dist_from_clusters > 1
                            )
            return is_interesting

    def estimate_earth_moving_cost(self, prod_dist, ref_dist, clusters):
        cost = 0
        for jdx in range(self.ref_dist.shape[1]):
            dirt_required = prod_dist[jdx] - ref_dist[jdx]
            dictn = []
            for kdx in range(self.clusters.shape[1]):
                if jdx == kdx:
                    dirt_can_be_transported = 0
                else:
                    dirt_can_be_transported = prod_dist[kdx] - ref_dist[kdx]
                this_dirt = -dirt_can_be_transported
                if this_dirt * dirt_required < 0:
                    cost_per_dirt = 1000000000
                else:
                    cost_per_dirt = np.sum(np.abs(clusters[kdx] - clusters[jdx]))
                dictn.append(
                    {
                        "dirt": np.abs(this_dirt),
                        "idx": kdx,
                        "cost_per_dirt": cost_per_dirt,
                    }
                )
            dictn = sorted(dictn, key=lambda x: x["cost_per_dirt"])
            this_cost = 0
            dirt_required = np.abs(dirt_required)
            for kdx in range(len(dictn)):
                if dirt_required > 0:
                    this_cost += (
                        min(dictn[kdx]["dirt"], dirt_required)
                        * dictn[kdx]["cost_per_dirt"]
                    )
                    dirt_required -= min(dictn[kdx]["dirt"], dirt_required)
                else:
                    break
            cost += this_cost
        return cost

    def bucket_reference_dataset(self):
        self.ref_dist = []
        self.prod_dist = []
        self.ref_dist_counts = []
        self.prod_dist_counts = []
        data = read_json(self.reference_dataset)
        all_inputs = np.array(
            [self.measurable.extract_val_from_training_data(x) for x in data]
        )
        all_inputs_shape = list(all_inputs.shape)
        if len(all_inputs_shape) == 1:
            all_inputs_shape.insert(1, 1)
            all_inputs = np.reshape(all_inputs, all_inputs_shape)

        if self.is_embedding:
            self.bucket_vector(all_inputs)
        else:
            buckets = []
            clusters = []
            cluster_vars = []
            for idx in range(all_inputs.shape[1]):
                this_inputs = all_inputs[:, idx]
                this_buckets, this_clusters, this_cluster_vars = self.bucket_scalar(
                    this_inputs
                )
                buckets.append(this_buckets)
                clusters.append(this_clusters)
                cluster_vars.append(this_cluster_vars)
            self.buckets = np.array(buckets)
            self.clusters = np.array(clusters)
            self.cluster_vars = np.array(cluster_vars)

        self.ref_dist = np.array(self.ref_dist)
        self.ref_dist_counts = np.array(self.ref_dist_counts)
        self.prod_dist = np.array(self.prod_dist)
        self.prod_dist_counts = np.array(self.prod_dist_counts)
        self.prod_dist_counts_arr.append(copy.deepcopy(self.prod_dist_counts))

    def bucket_scalar(self, arr):
        if isinstance(arr[0], str):
            uniques, counts = np.unique(np.array(arr), return_counts=True)
            buckets = uniques
            self.NUM_BUCKETS = len(buckets)
            clusters = uniques
            cluster_vars = [None] * self.NUM_BUCKETS
            self.ref_dist.append([[counts[x] / len(arr)] for x in range(self.NUM_BUCKETS)])
            self.ref_dist_counts.append(
                [[counts[x]] for x in range(self.NUM_BUCKETS)]
            )
        else:
            sorted_arr = np.sort(arr)
            buckets = []
            clusters = []
            cluster_vars = []
            for idx in range(0, self.NUM_BUCKETS):
                if idx > 0:
                    buckets.append(
                        sorted_arr[int(idx * (len(sorted_arr) - 1) / self.NUM_BUCKETS)]
                    )
                this_bucket_elems = sorted_arr[
                    int((idx) * (len(sorted_arr) - 1) / self.NUM_BUCKETS) : int(
                        (idx + 1) * (len(sorted_arr) - 1) / self.NUM_BUCKETS
                    )
                ]
                gaussian_mean = np.mean(this_bucket_elems)
                gaussian_var = np.var(this_bucket_elems)
                clusters.append([gaussian_mean])
                cluster_vars.append([gaussian_var])

            self.ref_dist.append([[1 / self.NUM_BUCKETS] for x in range(self.NUM_BUCKETS)])
            self.ref_dist_counts.append(
                [[int(len(sorted_arr) / self.NUM_BUCKETS)] for x in range(self.NUM_BUCKETS)]
            )
        self.prod_dist.append([[0] for x in range(self.NUM_BUCKETS)])
        self.prod_dist_counts.append([[0] for x in range(self.NUM_BUCKETS)])
        return np.array(buckets), np.array(clusters), np.array(cluster_vars)

    def bucket_vector(self, data):

        abs_data = np.abs(data)
        self.max_along_axis = np.max(abs_data, axis=0)
        data = data/self.max_along_axis

        plot_save_name = "training_dataset_clusters.png"
        # if self.cluster_plot_func is not None:
        #     self.log_handler.add_writer(self.dashboard_name)
        #     plot_save_name = self.log_handler.get_plot_save_name("training_dataset_clusters.png", self.dashboard_name)
        all_clusters, counts, cluster_vars = cluster_and_plot_data(
            data,
            self.NUM_BUCKETS,
            cluster_plot_func=self.cluster_plot_func,
            plot_save_name=plot_save_name,
            normalisation=self.max_along_axis
        )

        self.clusters = np.array([all_clusters])
        self.cluster_vars = np.array([cluster_vars])

        self.ref_dist_counts = np.array([counts])
        self.ref_dist = self.ref_dist_counts / data.shape[0]
        self.prod_dist = np.zeros((1, self.NUM_BUCKETS))
        self.prod_dist_counts = np.zeros((1, self.NUM_BUCKETS))



import numpy as np
from uptrain.core.classes.anomalies import AbstractAnomaly
from uptrain.constants import Anomaly


class CustomAnomaly(AbstractAnomaly):
    anomaly_type = Anomaly.CUSTOM_MONITOR

    def __init__(self, fw, check):
        self.dashboard_name = check.get("dashboard_name", "custom_measure")
        self.log_handler = fw.log_handler
        self.log_handler.add_writer(self.dashboard_name)
        self.check_func = check["check_func"]
        self.need_gt = check["need_gt"]

        initialize_func = check.get("initialize_func", None)
        if initialize_func is not None:
            initialize_func(self)

    def check(self, inputs, outputs, gts=None, extra_args={}):
        check_func = self.check_func
        return check_func(self, inputs, outputs, gts=gts, extra_args=extra_args)

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        return np.array([False] * len(extra_args["id"]))

    def need_ground_truth(self):
        return self.need_gt



import umap
from sklearn.cluster import DBSCAN
import numpy as np

from uptrain.core.classes.visuals import AbstractVisual
from uptrain.constants import Visual, Statistic
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.core.lib.helper_funcs import read_json


class Umap(AbstractVisual):
    visual_type = Visual.UMAP
    dashboard_name = "umap_and_clusters"

    def __init__(self, fw, check):
        self.framework = fw
        self.log_handler = fw.log_handler
        self.log_handler.add_writer(self.dashboard_name)
        self.measurable = MeasurableResolver(check.get("measurable_args", None)).resolve(fw)
        self.aggregate_measurable = MeasurableResolver(check.get("aggregate_args", None)).resolve(
            fw
        )
        self.label_measurable = MeasurableResolver(check.get("label_args", None)).resolve(
            fw
        )
        self.item_counts = {}
        self.feats_dictn = {}
        self.distances_dictn = {}
        self.count_checkpoints = check.get("count_checkpoints", ["all"])
        self.min_dist = check["min_dist"]
        self.n_neighbors = check["n_neighbors"]
        self.metric_umap = check["metric_umap"]
        self.dim = check["dim"]
        self.min_samples = check.get("min_samples", 5)
        self.eps = check.get("eps", 2)
        self.emb_dict = {}
        self.measurable_dict = {}
        self.total_count = 0
        self.umap_update_freq = check.get('umap_update_freq', 10000)
        self.vals = []
        self.labels = []
        self.do_clustering = check.get("do_clustering", False)
        
        self.initial_dataset = check.get('initial_dataset', None)
        if self.initial_dataset is not None:
            data = read_json(self.initial_dataset)
            self.vals.extend(
                [self.measurable.extract_val_from_training_data(x) for x in data]
            )
            if self.label_measurable is not None:
                self.labels.extend(
                    [self.label_measurable.extract_val_from_training_data(x) for x in data]
                )


    def check(self, inputs, outputs, gts=None, extra_args={}):
        if self.measurable is not None:
            vals = self.measurable.compute_and_log(
                inputs, outputs, gts=gts, extra=extra_args
            )
            self.vals.extend(vals)
            if self.label_measurable is not None:
                labels = self.label_measurable.compute_and_log(
                    inputs, outputs, gts=gts, extra=extra_args
                )
                self.labels.extend(labels)

        self.total_count += 1
        if not (self.total_count % self.umap_update_freq == 0):
            return

        for count in self.count_checkpoints:
            emb_list = np.array(self.get_data_for_umap(count))

            if emb_list.shape[0] > 10:
                # emb_list = np.squeeze(np.array(list(data_dict.values())))
                umap_list, clusters = self.get_umap_and_labels(
                    emb_list,
                    self.dim,
                    self.n_neighbors,
                    self.min_dist,
                    self.metric_umap,
                    self.eps,
                    self.min_samples,
                )
                this_data = {"umap": umap_list, "clusters": clusters}
                self.log_handler.add_histogram(
                    "umap_and_clusters",
                    this_data,
                    self.dashboard_name,
                    count,
                )

    def get_data_for_umap(self, count):
        if self.measurable is None:
            distribution_anomaly = list(
                filter(
                    lambda x: x.statistic_type == Statistic.DISTRIBUTION_STATS,
                    self.framework.check_manager.statistics_to_check,
                )
            )[0]
            return distribution_anomaly.get_feats_for_clustering(count)
        else:
            return self.vals

    def get_umap_and_labels(
        self,
        emb_list,
        dim,
        n_neighbors,
        min_dist,
        metric,
        eps,
        min_samples,
    ):
        if dim == "2D":
            n_components = 2
        else:
            n_components = 3

        umap_embeddings = umap.UMAP(
            n_neighbors=n_neighbors,
            n_components=n_components,
            min_dist=min_dist,
            metric=metric,
        ).fit_transform(emb_list)

        # Do DBSCAN clustering
        if self.do_clustering:
            clustering = DBSCAN(eps=eps, min_samples=min_samples).fit(umap_embeddings)
            labels = clustering.labels_
        else:
            labels = self.labels
        return umap_embeddings, labels



import os
from tensorboardX import SummaryWriter
import threading


class TensorboardLogs:
    def __init__(self, log_folder):
        self.log_folder = os.path.join(log_folder, "tb_data")
        os.makedirs(self.log_folder, exist_ok=True)
        launch_tb = lambda: os.system("tensorboard --logdir=" + self.log_folder)
        t = threading.Thread(target=launch_tb, args=([]))
        t.start()

    def add_writer(self, dashboard_name="UpTrain AI Logs"):
        return SummaryWriter(os.path.join(self.log_folder, dashboard_name))



import numpy as np

from uptrain.core.classes.anomalies import AbstractAnomaly
from uptrain.core.classes.algorithms import PopularityBias
from uptrain.constants import BiasAlgo, Anomaly


class RecommendationBias(AbstractAnomaly):
    dashboard_name = "popularity_bias"
    anomaly_type = Anomaly.POPULARITY_BIAS

    def __init__(self, fw, check):
        self.log_handler = fw.log_handler
        self.log_handler.add_writer(self.dashboard_name)
        self.acc_arr = []
        if check["algorithm"] == BiasAlgo.POPULARITY_BIAS:
            sessions = check.get("sessions", None)
            self.algo = PopularityBias(sessions)
        else:
            raise Exception("Recommendation bias type not supported")

    def need_ground_truth(self):
        return False

    def check(self, inputs, outputs, gts=None, extra_args={}):
        for y_pred in outputs:
            self.algo.add_prediction(y_pred)
            pop_arr = self.algo.all_popularity
            self.log_handler.add_histogram(
                "popularity_bias", pop_arr, self.dashboard_name
            )

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        return np.array([False] * len(extra_args["id"]))



import os
import shutil
import numpy as np


class LogHandler:
    def __init__(self, framework=None, cfg=None):
        self.fw = framework
        if os.path.exists(cfg.log_folder):
            print("Deleting the folder: ", cfg.log_folder)
            shutil.rmtree(cfg.log_folder)

        self.tb_writers = {}
        self.tb_logs = None
        if cfg.tb_logging:
            from uptrain.core.classes.logging.log_tensorboard import TensorboardLogs

            self.tb_logs = TensorboardLogs(cfg.log_folder)

        self.st_writer = None
        if cfg.st_logging:
            from uptrain.core.classes.logging.log_streamlit import StreamlitLogs

            self.st_log_folder = os.path.join(cfg.log_folder, "st_data")
            os.makedirs(self.st_log_folder, exist_ok=True)
            self.st_writer = StreamlitLogs(self.st_log_folder)
            self.st_log_folders_all = {}

    def add_writer(self, dashboard_name):
        dashboard_name = self.make_name_fold_directory_friendly(dashboard_name)
        tb_writer = None
        if self.tb_logs:
            tb_writer = self.tb_logs.add_writer(dashboard_name)
            self.tb_writers.update({dashboard_name: tb_writer})

    def get_plot_save_name(self, plot_name, dashboard_name):
        if self.st_writer:
            return os.path.join(self.st_log_folder, dashboard_name, plot_name)
        else:
            return ""

    def add_scalars(self, plot_name, dictn, count, dashboard_name):
        dashboard_name, plot_name = self.make_name_fold_directory_friendly(
            [dashboard_name, plot_name]
        )
        new_dictn = dict(
            zip(
                self.make_name_fold_directory_friendly(list(dictn.keys())),
                dictn.values(),
            )
        )
        if self.tb_logs:
            if dashboard_name in self.tb_writers:
                self.tb_writers[dashboard_name].add_scalars(plot_name, new_dictn, count)
        if self.st_writer:
            dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
            plot_folder = os.path.join(dashboard_dir, "line_plots", plot_name)
            os.makedirs(plot_folder, exist_ok=True)
            dictn.update({"count": count})
            self.st_writer.add_scalars(dictn, plot_folder)

    def add_histogram(self, plot_name, data, dashboard_name, count=-1):
        dashboard_name, plot_name = self.make_name_fold_directory_friendly(
            [dashboard_name, plot_name]
        )
        if not isinstance(data, dict):
            if isinstance(data, list) or isinstance(data, np.ndarray):
                if dashboard_name in self.tb_writers:
                    self.tb_writers[dashboard_name].add_histogram(plot_name, data, count)
        if self.st_writer:
            dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
            plot_folder = os.path.join(dashboard_dir, "histograms", plot_name)
            os.makedirs(plot_folder, exist_ok=True)
            self.st_writer.add_histogram(data, plot_folder, count)

    def add_alert(self, alert_name, alert, dashboard_name):
        # dashboard_name = self.make_name_fold_directory_friendly(
        #     [dashboard_name]
        # )
        if self.st_writer:
            dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
            plot_folder = os.path.join(dashboard_dir, "alerts")
            os.makedirs(plot_folder, exist_ok=True)
            self.st_writer.add_alert(alert_name, alert, plot_folder)

    def add_bar_graphs(self, plot_name, data, dashboard_name, count=-1):
        dashboard_name, plot_name = self.make_name_fold_directory_friendly(
            [dashboard_name, plot_name]
        )
        if self.st_writer:
            dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
            plot_folder = os.path.join(dashboard_dir, "bar_graphs", plot_name)
            os.makedirs(plot_folder, exist_ok=True)
            self.st_writer.add_bar_graphs(data, plot_folder, count)

    def make_name_fold_directory_friendly(self, arr):
        if isinstance(arr, str):
            return self.make_name_fold_directory_friendly([arr])[0]

        new_arr = [self.convert_str(x) for x in arr]
        return new_arr

    def convert_str(self, txt):
        txt = txt.replace("(", "_")
        txt = txt.replace(")", "_")
        txt = txt.replace(":", "_")
        txt = txt.replace(" ", "_")
        txt = txt.replace(",", "_")
        txt = txt.replace(">", "_")
        txt = txt.replace("<", "_")
        txt = txt.replace("=", "_")
        txt = txt.replace("-", "_")
        return txt
    
    def add_alert(self, alert_name, alert, dashboard_name):
        if self.st_writer:
            dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
            plot_folder = os.path.join(dashboard_dir, "alerts")
            os.makedirs(plot_folder, exist_ok=True)
            self.st_writer.add_alert(alert_name, alert, plot_folder)



import numpy as np

from uptrain.core.lib.helper_funcs import extract_data_points_from_batch
from uptrain.core.classes.distances import DistanceResolver
from uptrain.core.classes.statistics import AbstractStatistic
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Statistic


class Distribution(AbstractStatistic):
    dashboard_name = "distribution_stats"
    statistic_type = Statistic.DISTRIBUTION_STATS

    def __init__(self, fw, check):
        self.log_handler = fw.log_handler
        self.log_handler.add_writer(self.dashboard_name)
        self.measurable = MeasurableResolver(check["measurable_args"]).resolve(fw)
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.item_counts = {}
        self.feats_dictn = {}
        self.distances_dictn = {}
        self.count_checkpoints = check["count_checkpoints"]
        self.distance_types = check["distance_types"]
        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]

    def check(self, inputs, outputs, gts=None, extra_args={}):
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        updated_counts = []
        for idx in range(len(aggregate_ids)):
            if aggregate_ids[idx] not in self.item_counts:
                self.item_counts.update({aggregate_ids[idx]: 0})
            this_item_count = self.item_counts[aggregate_ids[idx]]
            if this_item_count in self.count_checkpoints:
                if this_item_count not in self.feats_dictn:
                    self.feats_dictn.update({this_item_count: {}})
                    self.distances_dictn.update({this_item_count: {}})
                this_val = extract_data_points_from_batch(vals, [idx])
                self.feats_dictn[this_item_count].update({aggregate_ids[idx]: this_val})
                if len(list(self.feats_dictn[this_item_count].keys())) > 1:
                    updated_counts.append(this_item_count)
                for agg_i in list(self.feats_dictn[this_item_count].keys()):
                    if agg_i == aggregate_ids[idx]:
                        continue
                    this_distances = dict(
                        zip(
                            self.distance_types,
                            [
                                x.compute_distance(
                                    this_val, self.feats_dictn[this_item_count][agg_i]
                                )
                                for x in self.dist_classes
                            ],
                        )
                    )
                    if len(self.distances_dictn[this_item_count]) == 0:
                        for distance_type in self.distance_types:
                            self.distances_dictn[this_item_count].update(
                                {distance_type: [this_distances[distance_type]]}
                            )
                    else:
                        for distance_key in list(this_distances.keys()):
                            self.distances_dictn[this_item_count][distance_key].append(
                                this_distances[distance_key]
                            )
            self.item_counts[aggregate_ids[idx]] += 1

        for count in list(self.distances_dictn.keys()):
            if count in updated_counts:
                for distance_type in self.distance_types:
                    plot_name = (
                        distance_type
                        # + " "
                        # + self.measurable.col_name()
                        # + " "
                        # + self.aggregate_measurable.col_name()
                    )

                    this_data = list(
                        np.reshape(
                            np.array(self.distances_dictn[count][distance_type]), -1
                        )
                    )
                    self.log_handler.add_histogram(
                        self.dashboard_name + "_" + plot_name,
                        this_data,
                        self.dashboard_name,
                        count,
                    )

    def get_feats_for_clustering(self, count):
        if count in self.feats_dictn:
            return self.feats_dictn[count]
        else:
            return {}



import collections

class PopularityBias:
    def __init__(self, sessions, k=10):
        self.k = k
        self.all_popularity = []

        # estimate popularity from sessions
        pop_map = collections.defaultdict(lambda : 0)
        num_interactions = 0
        for session in sessions:
            for event in session:
                pop_map[event] += 1
                num_interactions += 1

        # normalize popularity
        self.pop_map = {k:v/num_interactions for k,v in pop_map.items()}

    def add_prediction(self, y_pred):
        pop_one_pred = [self.pop_map.get(x, 0.0) for x in y_pred]
        self.all_popularity += pop_one_pred



import numpy as np
from copy import deepcopy

from uptrain.constants import Anomaly, Statistic, Visual, MeasurableType
from uptrain.core.classes.anomalies import (
    ConceptDrift,
    DataDrift,
    CustomAnomaly,
    RecommendationBias,
    DataIntegrity,
    EdgeCase,
)
from uptrain.core.classes.statistics import (
    Distance,
    Convergence,
    Distribution,
)
from uptrain.core.classes.visuals import Umap


class CheckManager:
    def __init__(self, framework, checks=[]):
        self.anomalies_to_check = []
        self.statistics_to_check = []
        self.visuals_to_check = []
        self.fw = framework
        for check in checks:
            if check["type"] in Anomaly:
                self.add_anomaly_to_monitor(check)
            if check["type"] in Statistic:
                self.add_statistics_to_monitor(check)
            if check["type"] in Visual:
                self.add_visuals(check)

    def add_anomaly_to_monitor(self, check):
        if check["type"] == Anomaly.EDGE_CASE:
            edge_case_manager = EdgeCase(self.fw, check["signal_formulae"])
            self.anomalies_to_check.append(edge_case_manager)
        elif check["type"] == Anomaly.CONCEPT_DRIFT:
            drift_manager = ConceptDrift(self.fw, check)
            self.anomalies_to_check.append(drift_manager)
        elif check["type"] == Anomaly.DATA_DRIFT:
            if "measurable_args" in check:
                drift_managers = [
                    DataDrift(
                        self.fw, check, is_embedding=check.get("is_embedding", None)
                    )
                ]
            else:
                drift_managers = []
                all_feats = self.fw.feat_name_list
                for feat in all_feats:
                    check_copy = deepcopy(check)
                    check_copy.update(
                        {
                            "measurable_args": {
                                "type": MeasurableType.INPUT_FEATURE,
                                "feature_name": feat,
                            }
                        }
                    )
                    drift_managers.append(
                        DataDrift(
                            self.fw,
                            check_copy,
                            is_embedding=check_copy.get("is_embedding", None),
                        )
                    )
            self.anomalies_to_check.extend(drift_managers)
        elif check["type"] == Anomaly.POPULARITY_BIAS:
            bias_manager = RecommendationBias(self.fw, check)
            self.anomalies_to_check.append(bias_manager)
        elif check["type"] == Anomaly.CUSTOM_MONITOR:
            custom_monitor = CustomAnomaly(self.fw, check)
            self.anomalies_to_check.append(custom_monitor)
        elif check["type"] == Anomaly.DATA_INTEGRITY:
            custom_monitor = DataIntegrity(self.fw, check)
            self.anomalies_to_check.append(custom_monitor)
        else:
            raise Exception("Anomaly type not Supported")

    def add_statistics_to_monitor(self, check):
        if check["type"] == Statistic.DISTANCE:
            custom_monitor = Distance(self.fw, check)
            self.statistics_to_check.append(custom_monitor)
        elif check["type"] == Statistic.DISTRIBUTION_STATS:
            custom_monitor = Distribution(self.fw, check)
            self.statistics_to_check.append(custom_monitor)
        elif check["type"] == Statistic.CONVERGENCE_STATS:
            custom_monitor = Convergence(self.fw, check)
            self.statistics_to_check.append(custom_monitor)
        else:
            raise Exception("Statistic type not Supported")

    def add_visuals(self, check):
        if check["type"] == Visual.UMAP:
            custom_monitor = Umap(self.fw, check)
            self.visuals_to_check.append(custom_monitor)
        else:
            raise Exception("Visual type not Supported")

    def check(self, inputs, outputs, gts=None, extra_args={}):
        for anomaly in self.anomalies_to_check:
            if anomaly.need_ground_truth() == (gts[0] is not None):
                anomaly.check(inputs, outputs, gts=gts, extra_args=extra_args)
        for stats in self.statistics_to_check:
            stats.check(inputs, outputs, gts=gts, extra_args=extra_args)
        for visuals in self.visuals_to_check:
            visuals.check(inputs, outputs, gts=gts, extra_args=extra_args)

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        is_interesting = []
        for anomaly in self.anomalies_to_check:
            if anomaly.need_ground_truth() == (gts[0] is not None):
                is_interesting.append(
                    anomaly.is_data_interesting(
                        inputs, outputs, gts=gts, extra_args=extra_args
                    )
                )
        return np.greater(np.sum(np.array(is_interesting), axis=0), 0)



import os
import csv
import threading
import json
import numpy as np


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class StreamlitLogs:
    def __init__(self, log_folder, log_every=10):
        self.log_every = log_every
        self.placeholders = {}
        self.prev_values = {}
        self.dfs = {}
        self.counts = {}
        self.log_folder = log_folder

        remote_st_py_file = "https://raw.githubusercontent.com/uptrain-ai/uptrain/dashboard/uptrain/core/classes/logging/st_run.py"
        # remote_st_py_file = "../../uptrain/core/classes/logging/st_run.py"

        cmd = "streamlit run " + remote_st_py_file + " -- " + self.log_folder
        launch_st = lambda: os.system(cmd)
        t = threading.Thread(target=launch_st, args=([]))
        t.start()

    def add_scalars(self, dict, folder):
        # CSV file that includes the data
        for key in dict.keys():
            if key == "count":
                continue
            file_name = os.path.join(folder, key + ".csv")
            if not os.path.isfile(file_name):
                with open(file_name, "w", newline="") as f_object:
                    writer = csv.writer(f_object)
                    writer.writerow([key, "count"])
                    f_object.close()

            with open(file_name, "a") as f_object:
                writer_object = csv.writer(f_object)
                writer_object.writerow([dict[key], dict["count"]])
                f_object.close()

    def add_histogram(self, data, folder, count=-1):
        file_name = os.path.join(folder, str(count) + ".json")
        with open(file_name, "w") as f:
            json.dump(data, f, cls=NumpyEncoder)

    def add_alert(self, alert_name, alert, folder):
        file_name = os.path.join(folder, str(alert_name) + ".json")
        with open(file_name, "w") as f:
            json.dump(alert, f)

    def add_bar_graphs(self, data, folder, count=-1):
        file_name = os.path.join(folder, str(count) + ".json")
        with open(file_name, "w") as f:
            json.dump(data, f, cls=NumpyEncoder)

    # def feat_slicing(self, fw):
    #     relevant_feat_list = st.sidebar.multiselect(
    #         "Select features", fw.feat_name_list, fw.feat_name_list[0]
    #     )
    #     scol1, scol2 = st.sidebar.columns(2)
    #     limit_list = []
    #     for feat in relevant_feat_list:
    #         with scol1:
    #             lower_limit = st.sidebar.slider(f"Lower limit for {feat}", value=0.5)
    #         with scol2:
    #             upper_limit = st.sidebar.slider(f"Upper limit for {feat}", value=0.55)
    #         limit_list.append((lower_limit, upper_limit))
    #     # button_callback = lambda : fw.feat_slicing(relevant_feat_list, limit_list)

    #     # TODO: Function is run from top every time the button is clicked
    #     button = st.sidebar.button(
    #         "Check",
    #         help="Check anomalies for this function",
    #     )
    #     # on_click=button_callback)
    #     if button:
    #         fw.feat_slicing(relevant_feat_list, limit_list)
